package com.bergerkiller.bukkit.nolagg.meta;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import net.minecraft.server.ChunkProviderServer;
import net.minecraft.server.WorldServer;

import org.bukkit.Chunk;
import org.bukkit.Location;
import org.bukkit.World;
import org.bukkit.craftbukkit.CraftWorld;
import org.bukkit.craftbukkit.entity.CraftEntity;
import org.bukkit.entity.Entity;
import org.bukkit.entity.ExperienceOrb;
import org.bukkit.entity.Item;

import com.all.bukkit.threadcheck.ThreadCheck;
import com.bergerkiller.bukkit.nolagg.ItemHandler;
import com.bergerkiller.bukkit.nolagg.SpawnHandler;
import com.bergerkiller.bukkit.nolagg.StackFormer;
import com.bergerkiller.bukkit.nolagg.Util;

public class WorldMetaData {
	private static final HashMap<UUID, WorldMetaData> worlddata = new HashMap<UUID, WorldMetaData>();
	
	public static void init() {
		for (WorldServer world : Util.getWorlds()) get(world);
	}
	public static void deinit() {
		for (WorldMetaData dat : worlddata.values()) {
			dat.remove();
		}
	}
	public static void remove(final World world) {
		if (!ThreadCheck.check(false)) return;
		WorldMetaData dat = worlddata.remove(world.getUID());
		if (dat != null) dat.remove();
	}
	public static WorldMetaData get(final World world) {
		if (!ThreadCheck.check(false)) return null;
		if (world == null) return null;
		WorldMetaData dat = worlddata.get(world.getUID());
		if (dat == null) {
			dat = new WorldMetaData(world);
			worlddata.put(world.getUID(), dat);
		}
		return dat;
	}
	public static WorldMetaData get(final net.minecraft.server.World world) {
		return get(world.getWorld());
	}
	public static void updateAll() {
		if (!ThreadCheck.check(false)) return;
		SpawnHandler.reset();
		for (WorldMetaData dat : worlddata.values()) dat.update();
	}
	public static int getTotalChunkCount() {
		int c = 0;
		for (WorldMetaData wmd : worlddata.values()) c += wmd.getChunkCount();	
		return c;
	}
	public static int getTotalUnloadableChunkCount() {
		int c = 0;
		for (WorldMetaData wmd : worlddata.values()) {
			for (ChunkMetaData cmd : wmd.getChunks()) {
			    if (cmd.isUnloadable()) c++;
			}
		}
		return c;
	}
	public static int getTotalHiddenItemCount() {
		int c = 0;
		for (WorldMetaData wmd : worlddata.values()) {
			for (ChunkMetaData cmd : wmd.getChunks()) {
			    c += cmd.getHiddenItemCount();
			}
		}
		return c;
	}

	public static void clearAllItems() {
		for (WorldMetaData dat : worlddata.values()) dat.clearItems();
	}
	public static boolean handleItemSpawn(Item item, boolean doIgnoreCheck) {
		return ChunkMetaData.get(item.getLocation()).handleItemSpawn(item, doIgnoreCheck);
	}
    public static boolean removeSpawnedItem(Item item) {
    	if (!ThreadCheck.check(false)) return false;
    	if (ItemHandler.maxItemsPerChunk <= 0) return false;
    	WorldMetaData dat = get(item.getWorld());
    	if (dat.getChunk(item.getLocation()).removeSpawnedItem(item)) {
    		return true;
    	} else {
    		//loop through chunks
    		for (ChunkMetaData cdat : dat.getChunks()) {
    			if (cdat.removeSpawnedItem(item)) return true;
    		}
    		return false;
    	}
    }
	
	private WorldMetaData(final World world) {
		this.world = (CraftWorld) world;
		this.chunkdata = new HashMap<Long, ChunkMetaData>();
		this.entitylist = new ArrayList<Entity>();
		this.itemlist = new ArrayList<Item>();
		this.orblist = new ArrayList<ExperienceOrb>();
		this.add();
	}
	
	public final CraftWorld world;
	private final Map<Long, ChunkMetaData> chunkdata;
	
	public ChunkProviderServer getChunkProvider() {
		return this.getNative().chunkProviderServer;
	}
	public WorldServer getNative() {
		return this.world.getHandle();
	}
	
	private final List<Entity> entitylist;
	private final List<Item> itemlist;
	private final List<ExperienceOrb> orblist;
	
	public int getHiddenItemCount() {
		if (!ThreadCheck.check(false)) return 0;
		int c = 0;
		for (ChunkMetaData cmd : this.getChunks()) c += cmd.getHiddenItemCount();
		return c;
	}
	public void clearItems() {
		if (!ThreadCheck.check(false)) return;
		for (ChunkMetaData dat : this.getChunks()) dat.clearItems();
	}

	public List<Entity> getEntities() {
		return this.entitylist;
	}
	public List<Item> getItems() {
		return this.itemlist;
	}
	public List<ExperienceOrb> getOrbs() {
		return this.orblist;
	}
	
	public Collection<ChunkMetaData> getChunks() {
		return this.chunkdata.values();
	}
	public ChunkMetaData getChunk(final Location location) {
		return this.getChunk(location.getBlockX() >> 4, location.getBlockZ() >> 4);
	}
	public ChunkMetaData getChunk(final net.minecraft.server.Chunk chunk) {
		return this.getChunk(chunk.x, chunk.z);
	}
 	public ChunkMetaData getChunk(final Chunk chunk) {
		return this.getChunk(chunk.getX(), chunk.getZ());
	}
	public ChunkMetaData getChunk(final int x, final int z) {
		if (!ThreadCheck.check(false)) return null;
		final long key = Util.toLong(x, z);
		ChunkMetaData dat = this.chunkdata.get(key);
		if (dat == null) {
			dat = new ChunkMetaData(this, x, z);
			this.chunkdata.put(key, dat);
		}
		return dat;
	}
	public int getChunkCount() {
		return this.chunkdata.size();
	}
	
	protected void removeChunk(final ChunkMetaData dat) {
		if (!ThreadCheck.check(false)) return;
		this.chunkdata.remove(Util.toLong(dat.getX(), dat.getZ()));
	}
	public boolean removeChunk(final int x, final int z) {
		if (!ThreadCheck.check(false)) return false;
		ChunkMetaData dat = this.chunkdata.get(Util.toLong(x, z));
		if (dat != null) {
			dat.remove();
			return true;
		} else {
			return false;
		}
	}
	
	private void update() {
		this.entitylist.clear();
		this.itemlist.clear();
		this.orblist.clear();
		CraftEntity e;
		for (Object obj : this.getNative().entityList) {
			e = (CraftEntity) ((net.minecraft.server.Entity) obj).getBukkitEntity();
			if (e.isDead()) {
				this.getNative().tracker.untrackEntity(e.getHandle());
				continue;
			}
			if (e instanceof Item) {
				if (ItemHandler.isIgnoredItem(e)) continue;
				this.itemlist.add((Item) e);
			} else if (e instanceof ExperienceOrb) {
				this.orblist.add((ExperienceOrb) e);
			}
			this.entitylist.add(e);
		}
		//update spawn limits
		SpawnHandler.update(this.getNative(), this.entitylist);

		//update item stacking
		StackFormer.update(this.itemlist, this.orblist);

		//update chunks and unload them if needed
		if (!this.chunkdata.isEmpty()) {
			final long time = System.currentTimeMillis() - ChunkMetaData.unloadDelay;
			Iterator<ChunkMetaData> it = this.chunkdata.values().iterator();
			ChunkMetaData dat;
			while (it.hasNext()) {
				dat = it.next();
				if (!dat.isLoaded()) {
					it.remove();
				} else if (!dat.isUnloadable() || dat.isPlayerNear()) {
					dat.update();
					dat.touch();
				} else if (dat.isExpired(time) && dat.unload()) {
					//unloaded
					it.remove();
				} else {
					dat.update();
				}
			}
		}
	}

	private void add() {
		for (Chunk c : world.getLoadedChunks()) getChunk(c);
	}
	private void remove() {
		Iterator<ChunkMetaData> it = this.chunkdata.values().iterator();
		ChunkMetaData dat;
		while (it.hasNext()) {
			dat = it.next();
			dat.remove(false);
			it.remove();
		}
		//other unload code here
	}

}
